
#include "stdafx.h"
#include <initguid.h>

#include "ArchiveExtractor.h"
#include "OpenCallback.h"
#include "ExtractCallback.h"
#include "Error.h"

#include "7zip/Common/RegisterArc.h"
#include "7zip/Common/FileStreams.h"
#include "7zip/Common/StreamObjects.h"
#include "7zip/UI/Common/LoadCodecs.h"
#include "7zip/UI/Common/OpenArchive.h"
#include "Common/StringConvert.h"

#include "Windows/PropVariant.h"
#include "Windows/PropVariantConversions.h"
#include "Windows/FileDir.h"

using NWindows::NCOM::CPropVariant;

CArchiveExtractor::CArchiveExtractor()
	: archive_(NULL)
	, archiveIndex_(ARCHIVE_NONE)
	, file_(NULL)
	, closeFileHandle_(false)
{
}

CArchiveExtractor::~CArchiveExtractor()
{
	// Release
	if (!fileList_.empty()) {
		for (FileList::iterator itr = fileList_.begin(); itr != fileList_.end(); ++itr) {
			delete *itr;
		}
	}

	// Don't close
	if (!closeFileHandle_ && (CInFileStream*)(file_) != NULL ) {
		file_->File.SetHandle(INVALID_HANDLE_VALUE);
	}
	archive_.Release();
}

bool CArchiveExtractor::Open(CCodecs *codecs, HANDLE hFile)
{
	try {
		// Check codecs
		if (codecs == NULL) {
			AM_SetError(ERROR_INVALID_ADDRESS);
			return false;
		}

		// Set Handle
		CInFileStream *fileSpec = new CInFileStream();
		CMyComPtr<CInFileStream> file = fileSpec;
		file->File.SetHandle(hFile);
		file_ = file;

		return Open(codecs, file);
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::OpenAs(CCodecs *codecs, HANDLE hFile, int index)
{
	try {
		// Check codecs
		if (codecs == NULL) {
			AM_SetError(ERROR_INVALID_ADDRESS);
			return false;
		}

		// Set Handle
		CInFileStream *fileSpec = new CInFileStream();
		CMyComPtr<CInFileStream> file = fileSpec;
		file->File.SetHandle(hFile);
		file_ = file;

		return OpenAs(codecs, file, index);
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::Open(CCodecs *codecs, const WCHAR *filename)
{
	try {
		// Check codecs
		if (codecs == NULL) {
			AM_SetError(ERROR_INVALID_ADDRESS);
			return false;
		}

		// Open file
		closeFileHandle_ = true;
		CInFileStream *fileSpec = new CInFileStream();
		CMyComPtr<CInFileStream> file = fileSpec;
		if (!file->Open(filename)) {
			AM_SetError(GetLastError());
			return false;
		}
		
		return Open(codecs, file);
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::OpenAs(CCodecs *codecs, const WCHAR *filename, int index)
{
	try {
		// Check codecs
		if (codecs == NULL) {
			AM_SetError(ERROR_INVALID_ADDRESS);
			return false;
		}

		// Open file
		closeFileHandle_ = true;
		CInFileStream *fileSpec = new CInFileStream();
		CMyComPtr<CInFileStream> file = fileSpec;
		if (!file->Open(filename)) {
			AM_SetError(GetLastError());
			return false;
		}
		
		return OpenAs(codecs, file, index);
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::Open(CCodecs *codecs, CMyComPtr<CInFileStream> file)
{
	try {
		// Open file as Archive
		CMyComPtr<IInArchive> archive;
		UString name, item;
		int idx;
		CArchiveOpenCallback *openCallbackSpec = new CArchiveOpenCallback();
		CMyComPtr<IArchiveOpenCallback> openCallback = openCallbackSpec;

		// Open
		HRESULT res = OpenArchive(codecs, file, name, &archive, idx, item, openCallbackSpec);
		if (res == S_OK ) {
			// OK
			archive_ = archive;
			archiveIndex_ = idx;
			if (!makeFileList()) {
				AM_SetError(GetLastError());
				return false;
			}
			AM_SetError(ERROR_SUCCESS);
			return true;
		}

		// Unspported format
		AM_SetError(ERROR_UNSUPPORTED_TYPE);
		return false;
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::OpenAs(CCodecs *codecs, CMyComPtr<CInFileStream> file, int index)
{
	try {
		// Open file as Archive
		CMyComPtr<IInArchive> archive;
		UString name, item;
		CArchiveOpenCallback *openCallbackSpec = new CArchiveOpenCallback();
		CMyComPtr<IArchiveOpenCallback> openCallback = openCallbackSpec;

		if (codecs->CreateInArchive(index, archive) == S_OK) {
			// Open
			HRESULT res = archive->Open(file, NULL, openCallbackSpec);
			if (res == S_OK ) {
				// OK
				archive_ = archive;
				archiveIndex_ = index;
				
				// Make file list
				if (!makeFileList()) {
					AM_SetError(ERROR_INVALID_DATA);
					return false;
				}

				AM_SetError(ERROR_SUCCESS);
				return true;
			}
		}

		// Unspported format
		AM_SetError(ERROR_UNSUPPORTED_TYPE);
		return false;
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}


bool CArchiveExtractor::ExtractToStream(UInt32 index, ISequentialOutStream *stream, ArchiveExtractCallback func, LPVOID param)
{
	try {
		if (fileList_.size() < index) {
			AM_SetError(ERROR_INVALID_INDEX);
			return false;
		}
		if (stream == NULL) {
			AM_SetError(ERROR_INVALID_ADDRESS);
			return false;
		}

		// Preparing Extract
		CExtractCallback *callbackSpec = new CExtractCallback();
		CMyComPtr<IArchiveExtractCallback> callback = callbackSpec;
		callbackSpec->Init(archive_);
		callbackSpec->OutStream_ = stream;

		// Set callback
		callbackSpec->ExtractCallbackParam_ = param;
		callbackSpec->ExtractCallback_ = func;
		
		UInt32 idx = fileList_[index]->Index;
		if (archive_->Extract(&idx, 1, 0, callbackSpec) == S_OK) {
			AM_SetError(ERROR_SUCCESS);
			return true;
		}

		// Set error code
		switch (callbackSpec->Result_) {
			case NArchive::NExtract::NOperationResult::kDataError:
				AM_SetError(ERROR_INVALID_DATA);
				break;
			case NArchive::NExtract::NOperationResult::kCRCError:
				AM_SetError(ERROR_CRC);
				break;
			case NArchive::NExtract::NOperationResult::kUnSupportedMethod:
				AM_SetError(ERROR_UNSUPPORTED_TYPE);
				break;
			default:
				// Win32 Error code
				AM_SetError(GetLastError());
				break;
		}
		return false;
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::ExtractToFile(UInt32 index, const WCHAR *file, ArchiveExtractCallback func, LPVOID param)
{
	try {
		if (fileList_.size() < index) {
			AM_SetError(ERROR_INVALID_INDEX);
			return false;
		}

		// Open file
		COutFileStream *outStreamSpec = new COutFileStream;
		CMyComPtr<ISequentialOutStream> outStream = outStreamSpec;
		if (!outStreamSpec->Open(file, CREATE_ALWAYS)) {
			AM_SetError(GetLastError());
			return false;
		}

		return ExtractToStream(index, outStreamSpec, func, param);
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::ExtractToMemory(UInt32 index, BYTE *buff, UInt32 buffSize, ArchiveExtractCallback func, LPVOID param)
{
	try {
		if (buff == NULL) {
			AM_SetError(ERROR_INVALID_ADDRESS);
			return false;
		}
		if (fileList_.size() < index) {
			AM_SetError(ERROR_INVALID_INDEX);
			return false;
		}

		// Memory
		CSequentialOutStreamImp2 *outStreamSpec = new CSequentialOutStreamImp2;
		CMyComPtr<ISequentialOutStream> outStream = outStreamSpec;
		outStreamSpec->Init(buff, buffSize);

		return ExtractToStream(index, outStreamSpec, func, param);
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::ExtractToHandle(UInt32 index, HANDLE handle, ArchiveExtractCallback func, LPVOID param)
{
	try {
		if (fileList_.size() < index) {
			AM_SetError(ERROR_INVALID_INDEX);
			return false;
		}

		// new COutFileStream
		COutFileStream *outStreamSpec = new COutFileStream;
		CMyComPtr<ISequentialOutStream> outStream = outStreamSpec;

		// Set handle
		outStreamSpec->SetHandle(handle);

		// Extract
		bool res = ExtractToStream(index, outStreamSpec, func, param);

		// Set handle
		outStreamSpec->SetHandle(INVALID_HANDLE_VALUE);
		
		return res;
	} catch (...) {
		AM_SetError(ERROR_OUTOFMEMORY);
		return false;
	}
}

bool CArchiveExtractor::makeFileList()
{
	// Get number of items(which contains folders and encrypted files)
	UInt32 numItems = 0;
	HRESULT res = archive_->GetNumberOfItems(&numItems);
	if (res != S_OK) {
		return false;
	}

	fileList_.clear();

	for (UInt32 i = 0; i < numItems; i++) {
		// FILTER: folder
		CPropVariant folder;
		HRESULT res = archive_->GetProperty(i, kpidIsFolder, &folder);
		if (res == S_OK && VARIANT_BOOLToBool(folder.boolVal)) {
			continue;
		}
		// FILTER: encrypted file
		CPropVariant encrypted;
		res = archive_->GetProperty(i, kpidEncrypted, &encrypted);
		if (res == S_OK && encrypted.vt == VT_BOOL && VARIANT_BOOLToBool(encrypted.boolVal)) {
			continue;
		}

		// Get path and size
		CPropVariant propPath, propSize;
		if (archive_->GetProperty(i, kpidPath, &propPath) != S_OK) {
			return false;
		}
		if (archive_->GetProperty(i, kpidSize, &propSize) != S_OK) {
			return false;
		}

		// Convert path to name
		UString path, name;
		path = ConvertPropVariantToString(propPath);
		NWindows::NFile::NDirectory::GetOnlyName(path, name);

		// Set
		FileInfo *info = new FileInfo;
		info->Index = i;
		info->Name = name;
		info->Path = path;
		info->Size = ConvertPropVariantToUInt64(propSize);
		fileList_.push_back(info);
	}

	return true;
}
